from pathlib import Path
import tempfile

import pytest

from mloq.files import File, test_req
from mloq.requirements import (
    compose_requirements,
    get_aliased_requirements_file,
    install_requirement_file,
    install_requirements,
    read_requirements_file,
    require_cuda,
    write_dev_requirements,
    write_project_requirements,
)


all_options = [
    "data-science",
    "datascience",
    "ds",
    "pytorch",
    "torch",
    "tensorflow",
    "tf",
    "data-visualization",
    "data-viz",
    "data-vis",
    "dataviz",
    "datavis",
]

requirements_all = ("data-science", "torch", "tf", "dataviz")


@pytest.fixture(params=all_options)
def option(request):
    return request.param


@pytest.fixture()
def all_requirements():
    return requirements_all


def test_get_aliased_requirements_file(option):
    val = get_aliased_requirements_file(option)
    assert isinstance(val, File)
    with pytest.raises(KeyError):
        get_aliased_requirements_file("miau")


def test_read_requirements_file(option):
    text = read_requirements_file(option)
    assert isinstance(text, str)


def test_compose_requirements(all_requirements):
    file = compose_requirements(all_requirements)
    lines = file.split("\n")
    test_file_path = Path(__file__).parent / "dummy_reqs.txt"
    with open(test_file_path, "r") as f:
        example_lines = f.read().split("\n")
    for line, example in zip(lines, example_lines):
        lib_name = line.split("==")[0]
        assert lib_name == example


def test_write_dev_requirements(ledger):
    with tempfile.TemporaryDirectory() as tmp:
        write_dev_requirements(ledger, tmp)
        write_dev_requirements(ledger, tmp, False, False, False)


def test_require_cuda():
    assert not require_cuda()
    assert not require_cuda(project_config={"requirements": "none"})
    assert not require_cuda(project_config={"requirements": "datascience"})
    assert require_cuda(project_config={"requirements": "torch"})


def test_write_project_requirements(ledger):
    assert write_project_requirements(["None"], ledger) is None


def test_install_requirements_file():
    result = install_requirement_file(test_req.src)
    assert result.ok


def test_install_requirements():
    install_requirements(".", test_req.src, test_req.src, test_req.src)



from pathlib import Path
import tempfile

import invoke
from omegaconf import DictConfig
import pytest

from mloq.api import setup_project, setup_requirements
from mloq.files import test_req


# race condition in /home/runner/.cache/flakehell which is created in the Git pre-commit hook
@pytest.mark.flaky(reruns=3)
def test_init_repository(config):
    with tempfile.TemporaryDirectory() as tmp:
        setup_project(
            path=Path(tmp),
            config=config,
            overwrite=False,
        )


def test_setup_requirements(ledger):
    try:
        setup_requirements(
            test_req.src.parent,
            config=DictConfig({"project": {}, "template": {}}),
            install=("all",),
            ledger=ledger,
        )
    except invoke.exceptions.UnexpectedExit:
        pass
    setup_requirements(
        test_req.src.parent,
        config=DictConfig({"project": {}, "template": {}}),
        install=("lint",),
        ledger=ledger,
    )



from pathlib import Path
import tempfile

from mloq.config.logic import write_config_setup


def test_write_config(mloq_yaml_dict):
    with tempfile.TemporaryDirectory() as tmp:
        write_config_setup(mloq_yaml_dict, path=Path(tmp) / "_mloq.yml")



"""This module defines common functionality for dealing with directories and files."""
import os
from pathlib import Path
from shutil import copyfile
from typing import Optional, Union

from omegaconf import DictConfig

from mloq import _logger
from mloq.failure import Failure
from mloq.files import File, init, Ledger, main, test_main, version
from mloq.templating import write_template


def copy_file(
    file: File,
    path: Union[Path, str],
    ledger: Ledger,
    overwrite: bool = False,
    description: Optional[str] = None,
) -> None:
    """
    Copy the file from src into dst.

    Args:
        file: File object representing the file that will be copied.
        path: Path to the destination of the copied file.
        ledger: Book keeper to keep track of the generated files.
        overwrite: If False, copy the file if it does not already exists in the \
                   target path. If True, overwrite the target file if it is already present.
        description: Description of the file that will be recorded by the Ledger.
    Returns:
        None.
    """
    target = path / file.dst
    if not os.path.isfile(str(target)) or overwrite:
        ledger.register(file, description=description)
        copyfile(file.src, target)
    else:
        _logger.debug(f"file {file.dst} already exists in {target}")


def create_project_skeleton(
    root_path: Path,
    config: DictConfig,
    ledger: Ledger,
    overwrite: bool = False,
) -> None:
    """
    Initialize the folder structure of a new Python project together with a bare minimum set of \
    code files.

    It creates a directory for the project containing an empty __init__.py file, and a \
    tests folder inside it with its corresponding __init__.py file. This function will only \
    create new files and folders in case they don't already exist, but it won't overwrite any \
    existing file.

    Args:
        root_path: Absolute path where the new project folder will be created.
        config: Contains all the parameters that define how the project will be set up.
        ledger: Book keeper for the generated files.
        overwrite: If False, copy the file if it does not already exists in the \
                   target path. If True, overwrite the target file if it is already present.

    Returns:
        None.
    """
    project_name = config.template.project_name
    # Project dir
    try:
        project_path = root_path / project_name
        os.makedirs(project_path, exist_ok=True)
        description = "Python package header for the project module"
        copy_file(init, project_path, ledger, overwrite, description=description)
        copy_file(version, project_path, ledger, overwrite)
        write_template(main, config=config, path=project_path, ledger=ledger, overwrite=overwrite)
        # Test dir inside project
        test_path = project_path / "tests"
        os.makedirs(test_path, exist_ok=True)
        write_template(
            test_main,
            config=config,
            path=test_path,
            ledger=ledger,
            overwrite=overwrite,
        )
        description = "Python package header for the test module"
        copy_file(init, test_path, ledger, overwrite, description=description)
    except (PermissionError, FileNotFoundError) as e:
        raise Failure() from e


def create_github_actions_directories(root_path: Union[str, Path]) -> None:
    """
    Initialize the folder structure for using GitHub actions workflows.

    It creates a .github directory on the target path, with an empty workflows \
    directory inside it, so the .github/workflows directory is ready to add GitHub \
    actions workflows. This function will only create new folders in case they don't already exist.

    Args:
        root_path: Absolute path where the .github folder will be crated.

    Returns:
        None.
    """
    gha_path = Path(root_path) / ".github"
    os.makedirs(gha_path, exist_ok=True)
    workflows_path = gha_path / "workflows"
    os.makedirs(workflows_path, exist_ok=True)


def create_docs_directories(root_path: Union[str, Path]) -> None:
    """
    Create the documentation directory structure.

    It creates a docs folder containing an images folder and a source folder. \
    The source folder contains an images folder, an _static folder, and a markdown folder.

    Args:
        root_path: Absolute path where the docs folder will be crated.

    Returns:
        None.
    """
    docs_path = Path(root_path) / "docs"
    os.makedirs(docs_path, exist_ok=True)
    imgs_path = docs_path / "images"
    os.makedirs(imgs_path, exist_ok=True)
    source_path = docs_path / "source"
    os.makedirs(source_path, exist_ok=True)
    static_path = source_path / "_static"
    os.makedirs(static_path, exist_ok=True)
    markdown_path = source_path / "markdown"
    os.makedirs(markdown_path, exist_ok=True)



"""Implements the `mloq setup` CLI command."""
import sys

import click
from omegaconf import DictConfig

from mloq.api import setup_project
from mloq.config.logic import get_docker_image, write_config_setup
from mloq.config.params import is_empty, PROJECT, TEMPLATE
from mloq.config.setup_generation import generate_config
from mloq.failure import Failure
from mloq.version import __version__


def generate_config_interactive(config: DictConfig) -> DictConfig:
    """Interactive generation of the project configuration."""
    # General project information
    click.echo("The following values will occur in several places in the generated files.")
    project, template = config.project, config.template
    template.project_name = TEMPLATE.project_name(template, True)
    template.description = TEMPLATE.description(template, True)
    template.owner = TEMPLATE.owner(template, True)
    template.email = TEMPLATE.email(template, True)
    template.author = TEMPLATE.author(template, True, default=template.owner)
    default_url = f"https://github.com/{template.owner}/{template.project_name}"
    template.project_url = TEMPLATE.project_url(template, True, default=default_url)
    click.echo()
    # License information
    click.echo("Please define the license of the project. ")
    click.echo(
        "Open Source projects will include the corresponding " "LICENSE file and a DCO.md file",
    )
    is_open_source = PROJECT.open_source(project, True, default=True)
    project.open_source = is_open_source
    default_license = "MIT" if is_open_source else "None"
    template.license = TEMPLATE.license(template, True, default=default_license)
    copyright_holder = TEMPLATE.copyright_holder(template, True, default=template.owner)
    template.copyright_holder = copyright_holder
    click.echo()
    # Python version and requirements
    click.echo(
        "What Python versions are supported? "
        "Please provide the values as a comma separated list. ",
    )
    versions = TEMPLATE.python_versions(template, True, default="3.6, 3.7, 3.8, 3.9")
    template.python_versions = versions
    click.echo("Please specify the requirements of the project as a comma separated list.")
    click.echo("Available values:")
    click.echo("    data-science: Common data science libraries such as numpy, pandas, sklearn...")
    click.echo(
        "    data-viz: Visualization libraries such as holoviews, bokeh, plotly, matplotlib...",
    )
    click.echo("    pytorch: Latest version of pytorch, torchvision and pytorch_lightning")
    click.echo("    tensorflow: ")  # , data-viz, torch, tensorflow}")
    project_requirements = PROJECT.requirements(project, True, default="None")
    project.requirements = project_requirements
    click.echo()
    # Continuous integration and other optional tools
    click.echo("You can configure the continuous integration using Github Actions.")
    click.echo("Available values:")
    click.echo("    Python: Push workflow for pure Python projects.")
    click.echo("    None: Do not set up the CI.")
    ci = PROJECT.ci(project, True, default="python")
    project.ci = ci
    if is_empty(ci):
        template.bot_name = "None"
        template.bot_email = "None"
    else:
        template.default_branch = TEMPLATE.default_branch(template, True, default="master")
        click.echo("A bot account will be used to automatically bump the version of your project.")
        template.bot_name = TEMPLATE.bot_name(template, True, default=template.author)
        template.bot_email = TEMPLATE.bot_email(template, True, default=template.email)

    template.ci_python_version = "3.8"
    template.ci_ubuntu_version = "ubuntu-20.04"
    template.ci_extra = ""
    template.pyproject_extra = ""
    template.docstring_checks = False
    project.docker = has_docker = PROJECT.docker(project, True, default=True)
    if has_docker:
        click.echo("MLOQ will generate a Dockerfile for your project.")
        base_docker = get_docker_image(config)
        if base_docker is not None:
            base_docker = TEMPLATE.docker_image(template, True, default=base_docker)
        template.docker_image = str(base_docker) if base_docker is None else base_docker
    project.docs = PROJECT.docs(project, True, default=True)
    click.echo("You can optionally create an ML Flow MLProject file.")
    project.mlflow = PROJECT.mlflow(project, True, default=False)
    project.git_init = git_init = PROJECT.git_init(project, True, default=True)
    if git_init:
        project.git_push = PROJECT.git_push(project, True, default=False)
        template.git_message = TEMPLATE.git_message(
            template,
            True,
            default="Generate project files with mloq",
        )
    return config


def welcome_message():
    """Welcome message to be displayed during interactive setup."""
    click.echo(f"Welcome to the MLOQ {__version__} interactive setup utility.")
    click.echo()
    click.echo(
        "Please enter values for the following settings (just press Enter "
        "to accept a default value, if one is given in brackets).",
    )
    click.echo()


def setup_cmd(
    config: DictConfig,
    output,
    overwrite: bool,
    interactive: bool,
    only_config: bool,
) -> int:
    """Initialize a new project using ML Ops Quickstart."""
    if interactive:
        welcome_message()
        config = generate_config_interactive(config)
    else:
        config = generate_config(config)
    if interactive and (only_config or click.confirm("Do you want to generate a mloq.yml file?")):
        write_config_setup(config, output, safe=True)
    if only_config:
        return 0
    if not overwrite and interactive:
        overwrite = click.confirm("Do you want to overwrite existing files?")
    try:
        setup_project(
            path=output,
            config=config,
            overwrite=overwrite,
        )
    except Failure as e:
        print(f"Failed to setup the project: {e}", file=sys.stderr)
        return 1
    return 0



"""Setup Git repository for the project."""


from pathlib import Path
import subprocess
from typing import Union

from omegaconf import DictConfig

from mloq.failure import Failure


def setup_git(
    path: Union[Path, str],
    config: DictConfig,
) -> None:
    """Initialize a Git repository over the generated files."""
    git_init = config.project.git_init
    if not git_init:
        return
    message = config.template.git_message
    push = config.project.git_push
    branch = config.template.default_branch
    project_name = config.template.project_name
    owner = config.template.owner
    sign_off = config.project.open_source
    path = str(path)
    try:
        _git_cmd(path, "init")
        _git_cmd(path, *f"remote add origin ssh://git@github.com/{owner}/{project_name}".split())
        subprocess.run(("pre-commit", "install"), check=True, cwd=path)

        def commit_all():
            _git_cmd(path, *"add .".split())
            _git_cmd(path, *f"commit {'--signoff' if sign_off else ''} -m".split(), message)

        if push:
            _git_cmd(path, *f"checkout -b {branch}".split())
            _git_cmd(path, *"add README.md".split())
            _git_cmd(
                path,
                *f"commit {'--signoff' if sign_off else ''} -m".split(),
                f"Initialize {branch}",
            )
            _git_cmd(path, *f"push origin {branch}".split())

        commit_all()

        if push:
            _git_cmd(path, *f"push origin HEAD:init-{branch}".split())
    except subprocess.CalledProcessError as e:
        raise Failure() from e


def _git_cmd(git_dir: str, *parts: str) -> None:
    subprocess.run(("git",) + parts, check=True, cwd=git_dir)



from pathlib import Path

import omegaconf
import pytest

from mloq.files import Ledger


@pytest.fixture(scope="module")
def project_config():
    test_project_config = {
        "open_source": True,
        "docker": True,
        "ci": "python",
        "mlflow": True,
        "docs": False,
        "requirements": ["torch"],
        "git_init": True,
        "git_push": False,
    }
    return test_project_config


@pytest.fixture(scope="module")
def template_config():
    test_template = {
        "project_name": "test_project",
        "default_branch": "master",
        "owner": "owner_test",
        "author": "owner_test",
        "email": "owner@mail.com",
        "copyright_year": 2021,
        "copyright_holder": "owner_test",
        "project_url": "https://github.com/owner_test/test_project",
        "bot_name": "owner_test",
        "bot_email": "owner@mail.com",
        "license": "MIT",
        "description": "test description of the project",
        "python_versions": ["3.6", "3.7", "3.8", "3.9"],
        "docker_image": "fragiletech/ubuntu18.04-cuda-11.0-py39",
        "ci_python_version": "3.8",
        "ci_ubuntu_version": "ubuntu-20.04",
        "ci_extra": "",
        "docstring_checks": False,
        "pyproject_extra": "",
        "git_message": "Test init commit",
    }
    return omegaconf.DictConfig(test_template)


@pytest.fixture(scope="module")
def config(project_config, template_config):
    return omegaconf.DictConfig({"project": project_config, "template": template_config})


@pytest.fixture(scope="module")
def mloq_yaml_dict():
    data = {
        "project": {
            "open_source": True,
            "docker": True,
            "ci": "python",
            "mlflow": False,
            "requirements": ["None"],
            "git_init": True,
            "git_push": False,
        },
        "template": {
            "project_name": "miau",
            "default_branch": "master",
            "owner": "guille",
            "author": "guille",
            "email": "maiol@maic.om",
            "copyright_holder": "guille",
            "project_url": "https://github.com/guille/miau",
            "bot_name": "guille",
            "bot_email": "maiol@maic.om",
            "license": "MIT",
            "description": "caca",
            "python_versions": ["3.6", "3.7", "3.8", "3.9"],
            "docker_image": "fragiletech/ubuntu18.04-base-py39",
            "git_message": "Generate project files with mloq",
        },
    }
    return data


@pytest.fixture()
def repo_path():
    return Path(__file__).parent / "test_project"


@pytest.fixture(scope="function")
def ledger():
    return Ledger()



"""Contains the functions that generate the required configuration."""
from datetime import datetime
from typing import Optional

from omegaconf import DictConfig

from mloq.config.logic import get_docker_image, load_empty_config_package
from mloq.config.params import PROJECT, TEMPLATE


def _generate_project_config(config: Optional[DictConfig] = None) -> DictConfig:
    """
    Generate a dictionary containing the necessary values to define the project root configuration.

    The generated config contains all the values defined under the project_config
    section in root.yml.

    Args:
        config: default values of the parameters.

    Returns:
        Dictionary containing the generated "project" config parameters.
    """
    project = load_empty_config_package().project
    project.update(config.project if config is not None else {})
    # Fill in the project settings
    project.open_source = PROJECT.open_source(project, False, default=True)
    project.docker = PROJECT.docker(project, False, default=True)
    project.mlflow = PROJECT.mlflow(project, False, default=False)
    return project


def _generate_template_config(config: Optional[DictConfig] = None) -> DictConfig:
    """
    Generate a dictionary containing the necessary values to customize the files generate by mloq.

    The generated config contains all the values defined under the template
    section in root.yml.

    Args:
        config: default values of the parameters.

    Returns:
        Dictionary containing the generated "template" config parameters.
    """
    # Initialize configuration placeholders
    if config is not None:
        _template, _project = config.template, config.project
    else:
        _template, _project = {}, {}
    config = load_empty_config_package()
    project, template = config.project, config.template
    project.update(_project)
    template.update(_template)
    # Fill in template values
    template.project_name = TEMPLATE.project_name(template, False)
    template.description = TEMPLATE.description(template, False)
    template.owner = TEMPLATE.owner(template, False)
    template.author = TEMPLATE.author(template, False, default=template.owner)
    template.email = TEMPLATE.email(template, False)
    copyright_holder = TEMPLATE.copyright_holder(template, False, default=template.owner)
    template.copyright_year = datetime.now().year
    template.copyright_holder = copyright_holder
    default_url = f"https://github.com/{template['owner']}/{template['project_name']}"
    template.project_url = TEMPLATE.project_url(template, False, default=default_url)
    versions = TEMPLATE.python_versions(template, False, default="3.6, 3.7, 3.8, 3.9")
    template.python_versions = versions
    # If project is undefined parse the full template
    if not project.open_source or project.open_source is None:
        template.license = "proprietary"
    else:
        template.license = TEMPLATE.license(template, False, default="MIT")
    if project.docker:
        base_docker = get_docker_image(config)
        if base_docker is not None:
            base_docker = TEMPLATE.docker_image(template, False, default=base_docker)
        template.docker_image = str(base_docker) if base_docker is None else base_docker
    return template


def generate_config(config: Optional[DictConfig] = None) -> DictConfig:
    """
    Generate a dictionary containing the necessary values to customize the files generated by mloq.

    Args:
        config: default values of the parameters.

    Returns:
        Dictionary containing the generated config parameters.
    """
    return DictConfig(
        {
            "project": _generate_project_config(config=config),
            "template": _generate_template_config(config=config),
        },
    )



"""The api module exposes the mloq functionality to the CLI."""
from pathlib import Path
from typing import List, Tuple, Union

from omegaconf import DictConfig

from mloq.files import (
    dockerfile,
    DOCS_FILES,
    Ledger,
    LICENSES,
    mlproject,
    OPEN_SOURCE_FILES,
    ROOT_PATH_FILES,
    setup_py,
    what_mloq_generated,
)
from mloq.git import setup_git
from mloq.requirements import install_requirements, write_requirements
from mloq.skeleton import create_docs_directories, create_project_skeleton
from mloq.templating import write_template
from mloq.workflows import setup_push_workflow


def setup_requirements(
    path: Union[Path, str],
    config: DictConfig,
    ledger: Ledger,
    lint: bool = True,
    test: bool = True,
    install: Union[str, Tuple[str], List[str], None] = None,
    overwrite: bool = False,
) -> None:
    """
    Write requirements files and install them if requested.

    This method writes the necessary requirement files (according to the
    selected options) of the new project. These files lists the exact
    versions of the packages needed to run the project. The method also
    allows installing the dependencies listed in the target requirements
    files.

    Args:
         path: Target folder where the configuration files will be written.
         config: DictConfig containing the selected project configuration.
         ledger: Ledger object. Keeps track of the generated files.
         lint: If True, writes the requirement-lint.txt file on the target
            folder.
         test: If True, writes the requirement-test.txt file on the target
            folder.
         install: Lists the target requirements files to be installed.
         overwrite: If True, overwrites existing files. Otherwise files
            that already exists will not be modified.
    """
    if isinstance(install, (tuple, list)) and "all" in install or install == "all":
        install_reqs = install_lint = install_test = True
    elif isinstance(install, (tuple, list)):
        install_reqs = "reqs" in install or "requirements" in install
        install_lint = "lint" in install
        install_test = "test" in install
    else:
        install_reqs = install_lint = install_test = False
    options = config.project.get("requirements", ["empty"])
    write_requirements(
        out_path=path,
        options=options,
        test=test,
        lint=lint,
        ledger=ledger,
        overwrite=overwrite,
    )
    install_requirements(
        path=path,
        requirements=install_reqs,
        test=install_test,
        lint=install_lint,
        py3=True,
    )


def setup_project_files(
    path: Union[Path, str],
    config: DictConfig,
    ledger: Ledger,
    overwrite: bool = False,
) -> None:
    """
    Write the template for common repository config files.

    This method initializes the structure of the new project, including
    primary code files. Additionally, it creates 'template'-related
    files (those concerning common repository configuration files) on the
    target folder.

    Args:
         path: Target folder where the generated files will be written.
         config: DictConfig containing the selected project configuration.
         ledger: Ledger object. Keeps track of the generated files.
         overwrite: If True, overwrites existing files. Otherwise files
            that already exists will not be modified.
    """
    path = Path(path)
    original_project_name = config.template.project_name
    config.template.project_name = config.template.project_name.replace("-", "_")
    try:
        create_project_skeleton(config=config, root_path=path, ledger=ledger, overwrite=overwrite)
        setup_root_files(
            config=config,
            path=path,
            overwrite=overwrite,
            ledger=ledger,
        )
    finally:
        config.template.project_name = original_project_name


def setup_root_files(
    path: Union[str, Path],
    config: DictConfig,
    ledger: Ledger,
    overwrite: bool = False,
) -> None:
    """
    Initialize root folder files.

    It writes repository files using as input the incoming DictConfig,
    which summarizes the options chosen by the user.

    Args:
         path: Target folder where the generated files will be written.
         config: DictConfig containing the selected project configuration.
         ledger: Ledger object. Keeps track of the generated files.
         overwrite: If True, overwrites existing files. Otherwise files
            that already exists will not be modified.
    """
    for file in ROOT_PATH_FILES:
        write_template(file, config=config, path=path, ledger=ledger, overwrite=overwrite)
    if config.project.get("docker"):
        write_template(dockerfile, config=config, path=path, ledger=ledger, overwrite=overwrite)
    if config.project.get("mlflow"):
        write_template(mlproject, config=config, path=path, ledger=ledger, overwrite=overwrite)
    if config.project.get("open_source"):
        for file in OPEN_SOURCE_FILES:
            write_template(file, config=config, path=path, ledger=ledger, overwrite=overwrite)
        license = config.template.license
        write_template(
            LICENSES[license],
            config=config,
            path=path,
            ledger=ledger,
            overwrite=overwrite,
        )
    else:
        license = "proprietary"
    license_classifiers = {
        "MIT": "License :: OSI Approved :: MIT License",
        "Apache-2.0": "License :: OSI Approved :: Apache Software License",
        "GPL-3.0": "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "proprietary": "License :: Other/Proprietary License",
    }
    license_classifer = license_classifiers[license]
    config = DictConfig({**config, "license_classifier": license_classifer})
    write_template(setup_py, config=config, path=path, ledger=ledger, overwrite=overwrite)


def dump_ledger(
    path: Union[str, Path],
    config: DictConfig,
    ledger: Ledger,
    overwrite: bool = False,
) -> None:
    """
    Write the summary of the generated files.

    This method collects the elements stored in ledger to create a markdown
    document that summarizes all the files generated by the MLOQ application.

    Args:
         path: Target folder where the generated files will be written.
         config: DictConfig containing the selected project configuration.
         ledger: Ledger object. Keeps track of the generated files.
         overwrite: If True, overwrites existing files. Otherwise files
            that already exists will not be modified.
    """
    config.template = DictConfig({**config.template, "generated_files": ledger.files})
    write_template(
        what_mloq_generated,
        config=config,
        path=path,
        ledger=ledger,
        overwrite=overwrite,
    )


def setup_docs(
    path: Union[str, Path],
    config: DictConfig,
    ledger: Ledger,
    overwrite: bool = False,
) -> None:
    """
    Configure the project documentation.

    When calling this function, it generates a documentation folder that
    contains necessary configuration files (requirement lists, 'make' commands,
    sphinx configurations) for building the documentation of the project.

    Args:
         path: Target folder where the generated files will be written.
         config: DictConfig containing the selected project configuration.
         ledger: Ledger object. Keeps track of the generated files.
         overwrite: If True, overwrites existing files. Otherwise files
            that already exists will not be modified.
    """
    if not config.project.get("docs", False):
        return
    create_docs_directories(root_path=path)
    source_files = {"conf.txt", "index.md"}
    docs_path = Path(path) / "docs"
    for file in DOCS_FILES:
        out_path = (docs_path / "source") if file.name in source_files else docs_path
        write_template(file, config=config, path=out_path, ledger=ledger, overwrite=overwrite)


def setup_project(
    path: Union[str, Path],
    config: DictConfig,
    overwrite: bool = False,
) -> None:
    """
    Initialize the project folder structure and all the filled-in boilerplate files.

    It calls the setup methods to generate (on the target path) the folder
    structure of the project, as well as repository files, requirement
    lists, and Github workflows (if requested).

    Args:
         path: Target folder where the generated files will be written.
         config: DictConfig containing the selected project configuration.
         overwrite: If True, overwrites existing files. Otherwise files
            that already exists will not be modified.
    """
    assert isinstance(config, DictConfig)
    path = Path(path)
    ledger = Ledger()
    setup_project_files(path=path, config=config, ledger=ledger, overwrite=overwrite)
    setup_push_workflow(path=path, config=config, ledger=ledger, overwrite=overwrite)
    setup_docs(path=path, config=config, ledger=ledger, overwrite=overwrite)
    setup_requirements(
        path=path,
        config=config,
        test=True,
        lint=True,
        ledger=ledger,
        overwrite=overwrite,
    )
    dump_ledger(path=path, config=config, ledger=ledger, overwrite=overwrite)
    setup_git(path=path, config=config)


def docs_project(
    path: Union[str, Path],
    config: DictConfig,
    overwrite: bool = False,
) -> None:
    """
    Initialize the project folder structure and all the documentation files.

    It calls the documentation setup method to generate (on the target
    path) the documentation files required by the project. This folder
    will contain configuration files (requirement lists, 'make' commands,
    sphinx configurations) needed for building the documentation of
    the project.

    Args:
         path: Target folder where the generated files will be written.
         config: DictConfig containing the selected project configuration.
         overwrite: If True, overwrites existing files. Otherwise files
            that already exists will not be modified.
    """
    assert isinstance(config, DictConfig)
    path = Path(path)
    ledger = Ledger()
    setup_docs(path=path, config=config, ledger=ledger, overwrite=overwrite)
    dump_ledger(path=path, config=config, ledger=ledger, overwrite=overwrite)


def package_project(
    path: Union[str, Path],
    config: DictConfig,
    overwrite: bool = False,
) -> None:
    """
    Initialize the project folder structure and repository files.

    It calls the setup project method to initialize the skeleton of the
    new project and create repository code files. Using 'config' as input
    dictionary, this method fills in core assets to generate primary files
    on the target path.

    Args:
        path: Target folder where the generated files will be written.
        config: DictConfig containing the selected project configuration.
        overwrite: If True, overwrites existing files. Otherwise files
            that already exists will not be modified.
    """
    assert isinstance(config, DictConfig)
    path = Path(path)
    ledger = Ledger()
    setup_project_files(path=path, config=config, ledger=ledger, overwrite=overwrite)
    dump_ledger(path=path, config=config, ledger=ledger, overwrite=overwrite)



"""
This is mostly a copy paste from \
https://github.com/pallets/click/blob/2fc486c880eda9fdb746ed8baa49416acab9ea6d/src/click/termui.py

Modified to allow prompt input that has a different color than the prompt text, while keeping
the color of the default prompt values the same as the prompt text color.
"""  # noqa: D400
import io

from click.exceptions import Abort, UsageError
from click.types import Choice, convert_type
from click.utils import echo, LazyFile


# The prompt functions to use.  The doc tools currently overwrite these
# functions to customize how they work.
visible_prompt_func = input

_ansi_reset_all = "\033[0m"


def hidden_prompt_func(prompt):
    """Input hidden text from the user."""
    import getpass

    return getpass.getpass(prompt)


def _build_prompt(text, suffix, show_default=False, default=None, show_choices=True, type=None):
    prompt_ = text
    if type is not None and show_choices and isinstance(type, Choice):
        prompt_ += f" ({', '.join(map(str, type.choices))})"
    if default is not None and show_default:
        prompt_ = f"{prompt_} [{_format_default(default)}]"
    return f"{prompt_}{suffix}"


def _format_default(default):
    if isinstance(default, (io.IOBase, LazyFile)) and hasattr(default, "name"):
        return default.name

    return default


def prompt(
    text,
    default=None,
    hide_input=False,
    confirmation_prompt=False,
    type=None,
    value_proc=None,
    prompt_suffix=": ",
    show_default=True,
    err=False,
    show_choices=True,
) -> None:
    """Prompts a user for input.  This is a convenience function that can \
    be used to prompt a user for input later.

    If the user aborts the input by sending a interrupt signal, this
    function will catch it and raise a :exc:`Abort` exception.

    .. versionadded:: 7.0
       Added the show_choices parameter.

    .. versionadded:: 6.0
       Added unicode support for cmd.exe on Windows.

    .. versionadded:: 4.0
       Added the `err` parameter.

    :param text: the text to show for the prompt.
    :param default: the default value to use if no input happens.  If this
                    is not given it will prompt until it's aborted.
    :param hide_input: if this is set to true then the input value will
                       be hidden.
    :param confirmation_prompt: asks for confirmation for the value.
    :param type: the type to use to check the value against.
    :param value_proc: if this parameter is provided it's a function that
                       is invoked instead of the type conversion to
                       convert a value.
    :param prompt_suffix: a suffix that should be added to the prompt.
    :param show_default: shows or hides the default value in the prompt.
    :param err: if set to true the file defaults to ``stderr`` instead of
                ``stdout``, the same as with echo.
    :param show_choices: Show or hide choices if the passed type is a Choice.
                         For example if type is a Choice of either day or week,
                         show_choices is true and text is "Group by" then the
                         prompt will be "Group by (day, week): ".
    :return: None
    """
    result = None

    def prompt_func(text):
        f = hidden_prompt_func if hide_input else visible_prompt_func
        try:
            # Write the prompt separately so that we get nice
            # coloring through colorama on Windows
            echo(f"{text}{_ansi_reset_all}", nl=False, err=err)
            return f("")
        except (KeyboardInterrupt, EOFError):
            # getpass doesn't print a newline if the user aborts input with ^C.
            # Allegedly this behavior is inherited from getpass(3).
            # A doc bug has been filed at https://bugs.python.org/issue24711
            if hide_input:
                echo(None, err=err)
            raise Abort()

    if value_proc is None:
        value_proc = convert_type(type, default)

    prompt_ = _build_prompt(text, prompt_suffix, show_default, default, show_choices, type)

    while 1:
        while 1:
            value = prompt_func(prompt_)
            if value:
                break
            elif default is not None:
                value = default
                break
        try:
            result = value_proc(value)
        except UsageError as e:
            if hide_input:
                echo("Error: the value you entered was invalid", err=err)
            else:
                echo(f"Error: {e.message}", err=err)  # noqa: B306
            continue
        if not confirmation_prompt:
            return result
        while 1:
            value2 = prompt_func("Repeat for confirmation: ")
            if value2:
                break
        if value == value2:
            return result
        echo("Error: the two entered values do not match", err=err)


def confirm(
    text,
    default=False,
    abort=False,
    prompt_suffix=": ",
    show_default=True,
    err=False,
) -> bool:
    """Prompts for confirmation (yes/no question).

    If the user aborts the input by sending a interrupt signal this
    function will catch it and raise a :exc:`Abort` exception.

    .. versionadded:: 4.0
       Added the `err` parameter.

    :param text: the question to ask.
    :param default: the default for the prompt.
    :param abort: if this is set to `True` a negative answer aborts the
                  exception by raising :exc:`Abort`.
    :param prompt_suffix: a suffix that should be added to the prompt.
    :param show_default: shows or hides the default value in the prompt.
    :param err: if set to true the file defaults to ``stderr`` instead of
                ``stdout``, the same as with echo.
    :return: User's decision.
    """
    prompt_ = _build_prompt(text, prompt_suffix, show_default, "Y/n" if default else "y/N")
    while 1:
        try:
            # Write the prompt separately so that we get nice
            # coloring through colorama on Windows
            echo(f"{prompt_}{_ansi_reset_all}", nl=False, err=err)
            value = visible_prompt_func("").lower().strip()
        except (KeyboardInterrupt, EOFError):
            raise Abort()
        if value in ("y", "yes"):
            rv = True
        elif value in ("n", "no"):
            rv = False
        elif value == "":
            rv = default
        else:
            echo("Error: invalid input", err=err)
            continue
        break
    if abort and not rv:
        raise Abort()
    return rv



"""This module defines common functionality for rendering and writing File templates."""
from datetime import datetime
from pathlib import Path
from typing import Any, Mapping, Union

from jinja2 import Environment, FileSystemLoader, select_autoescape
from omegaconf import DictConfig

from mloq import _logger
from mloq.files import File, Ledger, read_file, TEMPLATES_PATH, WORKFLOWS_PATH


jinja_env = Environment(
    loader=FileSystemLoader([str(TEMPLATES_PATH), str(WORKFLOWS_PATH)]),
    autoescape=select_autoescape(["html", "xml"]),
    keep_trailing_newline=True,
)


def render_template(file: File, kwargs: Mapping[str, Any]) -> str:
    """
    Render a jinja template with the provided parameter dict.

    Args:
        file: File object representing the jinja template that will be rendered.
        kwargs: Dictionary containing the parameters key and corresponding values \
                that will be used to render the template.

    Returns:
        String containing the rendered template.
    """
    if file.is_static:
        return read_file(file)
    jinja_template = jinja_env.get_template(str(file.name))
    jinja_template.globals["now"] = datetime.now
    return jinja_template.render(**kwargs)


def write_template(
    file: File,
    config: DictConfig,
    path: Union[Path, str],
    ledger: Ledger,
    overwrite: bool = False,
):
    """
    Create new file containing the rendered template found in source_path.

    Args:
        file: File object representing the jinja template that will be rendered.
        config: OmegaConf dictionary containing the parameters key and corresponding values \
                that will be used to render the templates.
        path: Absolute path to the folder where the file will be written.
        ledger: Book keeper to keep track of the generated files.
        overwrite: If False, copy the file if it does not already exists in the \
                   target path. If True, overwrite the target file if it is already present.

    Returns:
        None.
    """
    assert isinstance(config, DictConfig)
    path = Path(path)
    if not overwrite and (path / file.dst).exists():
        _logger.debug(f"file {file.name} already exists. Skipping")
        return

    kwargs = {**config["template"], "project": config.project}
    ledger.register(file)
    rendered = render_template(file, kwargs)
    with open(path / file.dst, "w") as f:
        f.write(rendered)



"""This module contains the functionality for parsing ad modifying the project configuration."""
from pathlib import Path
from typing import Union

from omegaconf import DictConfig, OmegaConf

from mloq.files import docs_yml, package_yml, setup_yml
from mloq.requirements import require_cuda


def get_docker_python_version(template: DictConfig) -> str:
    """Return the highest python version defined for the project."""
    max_version = list(sorted(template["python_versions"]))[-1]
    version = max_version.replace(".", "")
    return f"py{version}"


def get_docker_image(config: DictConfig) -> Union[str, None]:
    """
    Add to params the base docker container that will be used to define the project's container.

    If the dependencies require cuda the base image will be gpu friendly.
    """
    docker_is_false = not config.project.get("docker")
    if docker_is_false:  # Project does not require Docker
        return "empty"
    # If docker image is defined, return current value even if it's an empty image
    if config.template.get("docker_image") is not None:
        return config.template["docker_image"]
    # Define the appropriate FragileTech docker container as base image
    v = get_docker_python_version(config.template)
    ubuntu_v = "ubuntu20.04" if v == "py38" else "ubuntu18.04"
    image = (
        f"fragiletech/{ubuntu_v}-cuda-11.0-{v}"
        if require_cuda(config.project)
        else f"fragiletech/{ubuntu_v}-base-{v}"
    )
    return image


def write_config_setup(config: DictConfig, path: Union[Path, str], safe: bool = False):
    """Write setup config in a yaml file."""
    if safe:
        path = Path(path)
        path = path / setup_yml.dst if path.is_dir() else path
    with open(path, "w") as f:
        OmegaConf.save(config, f)


def load_empty_config_setup() -> DictConfig:
    """Return a dictionary containing all the MLOQ setup config values set to None."""
    return OmegaConf.load(setup_yml.src)


def write_config_docs(config: DictConfig, path: Union[Path, str], safe: bool = False):
    """Write docs config in a yaml file."""
    if safe:
        path = Path(path)
        path = path / docs_yml.dst if path.is_dir() else path
    with open(path, "w") as f:
        OmegaConf.save(config, f)


def load_empty_config_docs() -> DictConfig:
    """Return a dictionary containing all the MLOQ docs config values set to None."""
    return OmegaConf.load(docs_yml.src)


def write_config_package(config: DictConfig, path: Union[Path, str], safe: bool = False):
    """Write root config in a yaml file."""
    if safe:
        path = Path(path)
        path = path / package_yml.dst if path.is_dir() else path
    with open(path, "w") as f:
        OmegaConf.save(config, f)


def load_empty_config_package() -> DictConfig:
    """Return a dictionary containing all the MLOQ root config values set to None."""
    return OmegaConf.load(package_yml.src)



"""Command line interface for mloq."""

from pathlib import Path
import sys
from unittest.mock import patch

import click
import hydra
from omegaconf import DictConfig

from mloq.files import docs_yml, package_yml, setup_yml


overwrite_opt = click.option(
    "--overwrite/--no-overwrite",
    "-o/ ",
    default=False,
    show_default=True,
    help="Value indicating whether to overwrite existing files.",
)
config_file_opt = click.option(
    "--filename",
    "-f",
    "config_file",
    default=None,
    show_default=True,
    help="Name of the repository config file",
    type=click.Path(exists=True, file_okay=True, dir_okay=True, resolve_path=True),
)
only_config_opt = click.option(
    "--only-config/--everything",
    "-c/ ",
    default=False,
    show_default=True,
    help="Value indicating whether to not generate all the files except mloq.yml.",
)
output_directory_arg = click.argument(
    "output_directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
)

interactive_opt = click.option(
    "--interactive/--no-interactive",
    "-i/ ",
    default=False,
    show_default=True,
    help="If True the configuration values will be defined interactively on the command line.",
)


@click.group()
def cli():
    """Command line interface for ML Ops Quickstart."""
    pass


@cli.command(context_settings=dict(ignore_unknown_options=True))
@config_file_opt
@output_directory_arg
@overwrite_opt
@interactive_opt
@only_config_opt
@click.argument("hydra_args", nargs=-1, type=click.UNPROCESSED)
def setup(
    config_file: str,
    output_directory: str,
    overwrite: bool,
    only_config: bool,
    interactive: bool,
    hydra_args: str,
) -> None:
    """Entry point of `mloq setup`."""
    from mloq.cli.setup_cmd import setup_cmd

    config_file = Path(config_file) if config_file else setup_yml.src
    hydra_args = ["--config-dir", str(config_file.parent)] + list(hydra_args)
    config = DictConfig({})

    @hydra.main(config_name=config_file.name)
    def load_config(loaded_config: DictConfig):
        nonlocal config
        config = loaded_config

    with patch("sys.argv", [sys.argv[0]] + list(hydra_args)):
        load_config()

    exit(setup_cmd(config, output_directory, overwrite, interactive, only_config))


@cli.command(context_settings=dict(ignore_unknown_options=True))
@config_file_opt
@output_directory_arg
@overwrite_opt
@interactive_opt
@only_config_opt
@click.argument("hydra_args", nargs=-1, type=click.UNPROCESSED)
def docs(
    config_file: str,
    output_directory: str,
    overwrite: bool,
    only_config: bool,
    interactive: bool,
    hydra_args: str,
) -> None:
    """
    Entry point of `mloq docs`.

    Command line option of MLOQ. Generates the necessary documentation
    files for the project. Generated files:

    * 'docs' folder containing the documentation of the project, as well
    as the necessary tools to improve it.

    * Customizable commands for building the documentation (included in 'Makefile').

    * 'requirements-docs.txt' file containing the required libraries to
    generate the documentation.

    * 'conf.py' configuration file for sphinx and documentation plugins.

    * [Optional] 'mloq.yml' file describing the configuration options
    selected for this project.
    """
    from mloq.cli.docs_cmd import docs_cmd

    config_file = Path(config_file) if config_file else docs_yml.src
    hydra_args = ["--config-dir", str(config_file.parent)] + list(hydra_args)
    config = DictConfig({})

    @hydra.main(config_name=config_file.name)
    def load_config(loaded_config: DictConfig):
        nonlocal config
        config = loaded_config

    with patch("sys.argv", [sys.argv[0]] + list(hydra_args)):
        load_config()

    exit(docs_cmd(config, output_directory, overwrite, interactive, only_config))


@cli.command(context_settings=dict(ignore_unknown_options=True))
@config_file_opt
@output_directory_arg
@overwrite_opt
@interactive_opt
@only_config_opt
@click.argument("hydra_args", nargs=-1, type=click.UNPROCESSED)
def package(
    config_file: str,
    output_directory: str,
    overwrite: bool,
    only_config: bool,
    interactive: bool,
    hydra_args: str,
) -> None:
    """
    Entry point of `mloq package`.

    Command line option of MLOQ. Generates the necessary root files
    for the project. Generated files:

    * 'root' folder containing an empty __init__.py file, __main__.py
    executable file, and a test folder with its corresponding __init__.py
    file.

    * Project 'LICENSE' (choose from MIT, Apache-2.0, GPL-3.0, Proprietary,
    None)

    * Customizable commands for development (included in 'Makefile').

    * 'README.md' and 'README' documentation.

    * 'pyproject.toml' configuration file for various development tools.

    * [Optional] 'mloq.yml' file describing the configuration options
    selected for this project.

    * [Optional] 'Dockerfile' to install the project.

    * [Optional] 'MLProject' file defining MLFlows projects.
    """
    from mloq.cli.package_cmd import package_cmd

    config_file = Path(config_file) if config_file else package_yml.src
    hydra_args = ["--config-dir", str(config_file.parent)] + list(hydra_args)
    config = DictConfig({})

    @hydra.main(config_name=config_file.name)
    def load_config(loaded_config: DictConfig):
        nonlocal config
        config = loaded_config

    with patch("sys.argv", [sys.argv[0]] + list(hydra_args)):
        load_config()

    exit(package_cmd(config, output_directory, overwrite, interactive, only_config))



import tempfile

from omegaconf import DictConfig

from mloq.workflows import setup_push_workflow, setup_workflow_template


def test_setup_workflow_template(ledger):
    # test empty workflow
    with tempfile.TemporaryDirectory() as tmp:
        setup_workflow_template(
            root_path=tmp,
            config=DictConfig({"project": {}, "template": {}}),
            ledger=ledger,
        )
    # Test invalid name workflow
    setup_workflow_template(
        root_path=tmp,
        config=DictConfig({"project": dict(ci="miau_db"), "template": {}}),
        ledger=ledger,
    )


def test_setup_push_workflow(ledger):
    assert (
        setup_push_workflow(
            None,
            config=DictConfig({"project": {}, "template": {}}),
            ledger=ledger,
            overwrite=False,
        )
        is None
    )



"""Contains the functions that generate the required configuration."""
from datetime import datetime
from typing import Optional

from omegaconf import DictConfig

from mloq.config.logic import load_empty_config_docs
from mloq.config.params import PROJECT, TEMPLATE


def _generate_project_config(config: Optional[DictConfig] = None) -> DictConfig:
    """
    Generate a dictionary containing the necessary values to define the project doc configuration.

    The generated config contains all the values defined under the project_config
    section in docs.yml.

    Args:
        config: default values of the parameters.

    Returns:
        Dictionary containing the generated "project" config parameters.
    """
    project = load_empty_config_docs().project
    project.update(config.project if config is not None else {})
    # Fill in the project settings
    project.docs = PROJECT.docs(project, False, default=True)
    project.requirements = PROJECT.requirements(project, False, default="None")
    return project


def _generate_template_config(config: Optional[DictConfig] = None) -> DictConfig:
    """
    Generate a dictionary containing the necessary values to customize the files generated by mloq.

    The generated config contains all the values defined under the template
    section in docs.yml.

    Args:
        config: default values of the parameters.

    Returns:
        Dictionary containing the generated "template" config parameters.
    """
    # Initialize configuration placeholders
    if config is not None:
        _template, _project = config.template, config.project
    else:
        _template, _project = {}, {}
    config = load_empty_config_docs()
    project, template = config.project, config.template
    project.update(_project)
    template.update(_template)
    # Fill in template values
    template.project_name = TEMPLATE.project_name(template, False)
    template.description = TEMPLATE.description(template, False)
    template.author = TEMPLATE.author(template, False)
    copyright_holder = TEMPLATE.copyright_holder(template, False, default=template.author)
    template.copyright_year = datetime.now().year
    template.copyright_holder = copyright_holder
    return template


def generate_config(config: Optional[DictConfig] = None) -> DictConfig:
    """
    Generate a dictionary containing the necessary values to customize the files generated by mloq.

    Args:
        config: default values of the parameters.

    Returns:
        Dictionary containing the generated config parameters.
    """
    return DictConfig(
        {
            "project": _generate_project_config(config=config),
            "template": _generate_template_config(config=config),
        },
    )



import copy
import os

from mloq.config.setup_generation import (
    _generate_project_config,
    _generate_template_config,
    generate_config,
)


def compare_dicts(a, b):
    for (k1, v1), (k2, v2) in zip(sorted(a.items()), sorted(b.items())):
        assert k1 == k2
        if not isinstance(v1, list):
            assert v1 == v2, k1
        else:
            for x1, x2 in zip(v1, v2):
                assert x1 == x2, k1


def test_generate_project_config(config):
    _config = copy.deepcopy(config)
    _config.project.requirements = None
    os.environ["MLOQ_REQUIREMENTS"] = "torch"
    try:
        new_project = _generate_project_config(_config)
        compare_dicts(config.project, new_project)
    finally:
        del os.environ["MLOQ_REQUIREMENTS"]


def test_generate_template_config(config):
    _config = copy.deepcopy(config)
    _config.template.project_name = None
    os.environ["MLOQ_PROJECT_NAME"] = "test_project"
    try:
        new_template = _generate_template_config(_config)
        compare_dicts(config.template, new_template)
    finally:
        del os.environ["MLOQ_PROJECT_NAME"]


def test_generate_config(config):
    _config = copy.deepcopy(config)
    _config.project.requirements = None
    _config.template.project_name = None
    os.environ["MLOQ_REQUIREMENTS"] = "torch"
    os.environ["MLOQ_PROJECT_NAME"] = "test_project"
    try:
        new_config = generate_config(_config)
        compare_dicts(config, new_config)
    finally:
        del os.environ["MLOQ_REQUIREMENTS"]
        del os.environ["MLOQ_PROJECT_NAME"]



"""This module defines all the different assets accessible from mloq."""
from pathlib import Path
import sys
from typing import List, NamedTuple, Optional, Tuple, Union


class File(NamedTuple):
    """Generated project file."""

    name: str
    src: Path
    dst: Path
    description: str
    is_static: bool


def file(
    name: str,
    path: Union[Path, str],
    description: Optional[str] = None,
    dst: Optional[Union[Path, str]] = None,
    is_static: bool = False,
) -> File:
    """Define a new asset as a File namedtuple."""
    if description is None:
        print("FIXME: %s must have a description" % name, file=sys.stderr)
        description = "TODO"
    dst = Path(dst) if dst is not None else name
    return File(
        name=name,
        src=Path(path) / name,
        dst=dst,
        description=description,
        is_static=is_static,
    )


class Ledger:
    """Keep track of the generated files."""

    def __init__(self):
        """Initialize a new instance of the Ledger class."""
        self._files = []

    @property
    def files(self) -> List[Tuple[str, str]]:
        """Return the list of generated file names."""
        return [(str(f), d) for f, d in sorted(self._files)]

    def register(self, file: Union[File, str, Path], description: Optional[str] = None) -> None:
        """Append another generated file to the book."""
        if isinstance(file, File):
            description = file.description if description is None else description
            file = file.dst
        else:
            assert description is not None
        self._files.append((Path(file), description))


# Assets paths
ASSETS_PATH = Path(__file__).parent / "assets"
CONFIGS_PATH = ASSETS_PATH / "configs"
TEMPLATES_PATH = ASSETS_PATH / "templates"
STATIC_FILES_PATH = ASSETS_PATH / "static"
REQUIREMENTS_PATH = ASSETS_PATH / "requirements"
WORKFLOWS_PATH = ASSETS_PATH / "workflows"

# Common files
setup_yml = file(
    "setup.yml",
    CONFIGS_PATH,
    "mloq configuration, you can safely remove it if you don't plan to upgrade",
    dst="mloq.yml",
    is_static=True,
)
docs_yml = file(
    "docs.yml",
    CONFIGS_PATH,
    "mloq configuration, you can safely remove it if you don't plan to upgrade",
    dst="mloq.yml",
    is_static=True,
)
package_yml = file(
    "package.yml",
    CONFIGS_PATH,
    "mloq configuration, you can safely remove it if you don't plan to upgrade",
    dst="mloq.yml",
    is_static=True,
)
gitignore = file(
    ".gitignore",
    STATIC_FILES_PATH,
    "list of files and directories ignored by Git operations",
    is_static=True,
)
pre_commit_hook = file(
    ".pre-commit-config.yaml",
    STATIC_FILES_PATH,
    "Git pre-commit hooks configuration",
    is_static=True,
)
dco = file(
    "DCO.md",
    STATIC_FILES_PATH,
    "Developer Certificate of Origin - needed in open source projects to certify that "
    "the incoming contributions are legitimate",
    is_static=True,
)
init = file(
    "init.txt",
    STATIC_FILES_PATH,
    "Python package header",
    dst="__init__.py",
    is_static=True,
)
main = file(
    "main.txt",
    STATIC_FILES_PATH,
    "Python package executable entry point",
    dst="__main__.py",
    is_static=True,
)
test_main = file(
    "test_main.txt",
    TEMPLATES_PATH,
    "Unit test of the python package executable entry point",
    dst="test_main.py",
    is_static=False,
)
version = file(
    "version.txt",
    STATIC_FILES_PATH,
    "defines the version of the package that is incremented on each push",
    dst="version.py",
    is_static=True,
)
code_of_conduct = file(
    "CODE_OF_CONDUCT.md",
    TEMPLATES_PATH,
    "behavioral rules and norms in open source projects",
)

# Requirements files
data_science_req = file(
    "data-science.txt",
    REQUIREMENTS_PATH,
    "list of commonly used data science libraries",
)
data_viz_req = file(
    "data-visualization.txt",
    REQUIREMENTS_PATH,
    "list of commonly used visualization libraries",
    is_static=True,
)
pytorch_req = file(
    "pytorch.txt",
    REQUIREMENTS_PATH,
    "Pytorch deep learning libraries",
    is_static=True,
)
tensorflow_req = file(
    "tensorflow.txt",
    REQUIREMENTS_PATH,
    "Tensorflow deep learning libraries",
    is_static=True,
)
lint_req = file(
    "requirements-lint.txt",
    REQUIREMENTS_PATH,
    "list of exact versions of the packages used to check your code style",
    is_static=True,
)
test_req = file(
    "requirements-test.txt",
    REQUIREMENTS_PATH,
    "list of exact versions of the packages needed to run your test suite",
    is_static=True,
)
dogfood_req = file(
    "dogfood.txt",
    REQUIREMENTS_PATH,
    "list of mock requirements for testing purposes",
    is_static=True,
)
docs_req = file(
    "requirements-docs.txt",
    REQUIREMENTS_PATH,
    "list of exact versions of the packages needed to build your documentation",
    is_static=True,
)
# Documentation files
makefile_docs = file(
    "makefile_docs.txt",
    STATIC_FILES_PATH,
    "common make commands for building the documentation",
    dst="Makefile",
    is_static=True,
)
make_bat_docs = file(
    "make_bat.txt",
    STATIC_FILES_PATH,
    "common make commands for building the documentation",
    dst="make.bat",
    is_static=True,
)
conf_py = file(
    "conf.txt",
    TEMPLATES_PATH,
    "configuration file for sphinx and doc plugins",
    dst="conf.py",
)
index_md = file(
    "index.md",
    TEMPLATES_PATH,
    "configuration file for sphinx and doc plugins",
)
# Templates
mit_license = file("MIT_LICENSE", TEMPLATES_PATH, "license of the project", dst="LICENSE")
apache_license = file("APACHE_LICENSE", TEMPLATES_PATH, "license of the project", dst="LICENSE")
gpl_license = file(
    "GPL_LICENSE",
    STATIC_FILES_PATH,
    "license of the project",
    dst="LICENSE",
    is_static=True,
)
setup_py = file(
    "setup.txt",
    TEMPLATES_PATH,
    "Python package installation metadata",
    dst="setup.py",
)
dockerfile = file("Dockerfile", TEMPLATES_PATH, "Dockerfile to install the project")
makefile = file("Makefile", TEMPLATES_PATH, "common make commands for development")
mlproject = file("MLproject", TEMPLATES_PATH, "file defining MLFlow projects")
readme = file("README.md", TEMPLATES_PATH, "README")
contributing = file(
    "CONTRIBUTING.md",
    TEMPLATES_PATH,
    "technical manual on how to contrib to the open source project",
)
pyproject_toml = file(
    "pyproject.toml",
    TEMPLATES_PATH,
    "configuration of various development tools: linters, formatters",
)
codecov = file(
    ".codecov.yml",
    TEMPLATES_PATH,
    "configuration of CodeCov service to track the code coverage",
)
what_mloq_generated = file("WHAT_MLOQ_GENERATED.md", TEMPLATES_PATH, "this file")

# Workflows
push_python_wkf = file(
    "push.yml",
    WORKFLOWS_PATH,
    "GitHub Actions continuous integration workflow file",
)

ROOT_PATH_FILES = [
    gitignore,
    pre_commit_hook,
    pyproject_toml,
    makefile,
    readme,
    codecov,
]

OPEN_SOURCE_FILES = [dco, contributing, code_of_conduct]

LICENSES = {
    "MIT": mit_license,
    "Apache-2.0": apache_license,
    "GPL-3.0": gpl_license,
}

WORKFLOW_FILES = [push_python_wkf]

PYTHON_FILES = [init, main, test_main, version]

DOCS_FILES = [conf_py, index_md, makefile_docs, make_bat_docs, docs_req]

REQUIREMENTS_FILES = [
    lint_req,
    pytorch_req,
    tensorflow_req,
    data_viz_req,
    data_science_req,
    test_req,
]


ALL_FILES = (
    ROOT_PATH_FILES
    + WORKFLOW_FILES
    + PYTHON_FILES
    + [setup_yml, docs_yml, package_yml]
    + DOCS_FILES
)

ALL_FILE_PATHS = [str(f.src) for f in ALL_FILES]


def read_file(file: File) -> str:
    """Return and string with the content of the provided file."""
    with open(file.src, "r") as f:
        return f.read()



"""This file contains the logic defining all the parameters needed to se up a project with mloq."""
import os
from typing import Any, List, NamedTuple, Optional, Set, Tuple, Union

import click
from omegaconf import DictConfig

from mloq.cli.custom_prompt import confirm, prompt


Choices = Union[List[str], Tuple[str], Set[str]]


class ConfigParam:
    """
    Defines a configuration parameter.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yml
        3. Interactive promp from CLI (Optional)
    """

    def __init__(self, name: str, text: Optional[str] = None, **kwargs):
        """
        Initialize a ConfigParam.

        Args:
            name: Name of the parameter (as defined in mloq.yml).
            text: Text that will be prompted in the CLI when using interactive mode.
            **kwargs: Passed to click.prompt when running in interactive mode.
        """
        self.name = name
        prompt_text = text if text is not None else self.name
        self._prompt_text = click.style(f"> {prompt_text}", fg="bright_magenta", reset=False)
        self._prompt_kwargs = kwargs
        self._prompt_kwargs["show_default"] = kwargs.get("show_default", True)
        self._prompt_kwargs["type"] = kwargs.get("type", str)

    def __call__(
        self,
        config: Optional[DictConfig] = None,
        interactive: bool = False,
        default: Optional[Any] = None,
        raise_error: bool = True,
        **kwargs,
    ):
        """
        Return the value of the parameter parsing it from the different input sources available.

        Args:
            config: Dictionary containing the configuration values defined in mloq.yml.
            interactive: Prompt the user to input the value from CLI if it's not defined \
                        in config or as en environment variable.
            default: Default value displayed in the interactive mode.
            raise_error: If True, a ValueError will be raise if the value is not defined. \
                         If False, return None when the value is not defined.
            **kwargs: Passed to click.prompt in interactive mode. Overrides the \
                      values defined in __init__

        Returns:
            Value of the parameter.
        """
        config = config or {}
        value = self._value_from_env()
        value = value if value is not None else self._value_from_config(config)
        value = value if value is not None else default
        if interactive:
            value = self._prompt(value, **kwargs)
        if value is None and raise_error:
            raise ValueError(f"Config value {self.name} is not defined.")
        return value

    def _value_from_env(self):
        """Return the config value if it is defined as an environment variable."""
        env_name = f"MLOQ_{self.name.upper()}"
        return os.environ.get(env_name)

    def _prompt(self, value, **kwargs):
        """Prompt user for value."""
        _kwargs = dict(self._prompt_kwargs)
        _kwargs.update(kwargs)
        if value is not None:
            _kwargs["default"] = value
        return prompt(self._prompt_text, **_kwargs)

    def _value_from_config(self, config: DictConfig):
        """Return the config value if it is present on the config dict."""

        def find_value(conf):
            for k, v in conf.items():
                if k == self.name:
                    return v
                elif isinstance(v, dict):
                    return find_value(v)

        return find_value(config)


class MultiChoiceParam(ConfigParam):
    """
    Define a configuration parameter that can take multiple values \
    from a pre-defined set of values.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yml
        3. Interactive promp from CLI (Optional)
    """

    def __init__(
        self,
        name: str,
        choices: Optional[Choices] = None,
        text: Optional[str] = None,
        **kwargs,
    ):
        """
        Initialize a ConfigParam.

        Args:
            name: Name of the parameter (as defined in mloq.yml).
            choices: Contains all the available values for the parameter.
            text: Text that will be prompted in the CLI when using interactive mode.
            **kwargs: Passed to click.prompt when running in interactive mode.
        """
        kwargs["type"] = str
        super(MultiChoiceParam, self).__init__(name=name, text=text, **kwargs)
        self.choices = choices  # TODO: use this to validate user input.

    def _prompt(self, value, **kwargs) -> List[str]:
        """Transform the parsed string from the CLI into a list of selected values."""
        val = super(MultiChoiceParam, self)._prompt(value, **kwargs)
        return self._parse_string(val) if isinstance(val, str) else val

    def _value_from_env(self):
        """Return the config value if it is defined as an environment variable."""
        val = super(MultiChoiceParam, self)._value_from_env()
        return self._parse_string(val) if isinstance(val, str) else val

    @staticmethod
    def _parse_string(value) -> List[str]:
        def filter_str(s):
            return s.lstrip().replace("'", "").replace('"', "").replace("[", "").replace("]", "")

        return [filter_str(s) for s in value.split(",")]


class BooleanParam(ConfigParam):
    """
    Defines a boolean configuration parameter.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yml
        3. Interactive promp from CLI (Optional)
    """

    def _prompt(self, value, **kwargs):
        """Prompt user for value."""
        _kwargs = dict(self._prompt_kwargs)
        _kwargs.update(kwargs)
        if "type" in _kwargs:
            del _kwargs["type"]
        if value is not None:
            _kwargs["default"] = value
        return confirm(self._prompt_text, **_kwargs)


EMPTY_VALUES = {"none", "empty", "skip"}


def is_empty(value: Union[str, list, tuple, set]) -> bool:
    """
    Return True if a the target value is defined as an empty value.

    Empty values are defined in the configuration (are not None) but they are not \
    used by mloq. For example, empty values can be used to define empty project \
    requirements or no base Docker image.
    """
    if value is None:
        return True
    elif isinstance(value, str):
        return value.lower() in EMPTY_VALUES
    elif not len(value):
        return True
    return any([v.lower() in EMPTY_VALUES for v in value])


"""Contains all the parameters that define how the project will be set up."""
_PROJECT = [
    MultiChoiceParam(
        "requirements",
        text="Project requirements",
        choices=["data-science", "data-viz", "torch", "tensorflow", "none"],
    ),
    BooleanParam("open_source", "Is the project Open Source?"),
    BooleanParam("docker", "Do you want to set up a Docker container with your project?"),
    BooleanParam("docs", "Do you want to set up the project documentation?"),
    ConfigParam(
        "ci",
        "Github Actions push workflow",
        type=click.Choice(["python", "none"], case_sensitive=False),
    ),
    BooleanParam("mlflow", "Do you want to set up ML Flow?"),
    BooleanParam("git_init", "Initialize Git repository?"),
    BooleanParam("git_push", "Execute git push to the target repository?"),
]

"""Contains all the parameters that are used to customize the generated template files."""
_TEMPLATE = [
    ConfigParam("project_name", "Select project name"),
    ConfigParam("owner", "Github handle of the project owner"),
    ConfigParam("email", "Owner contact email"),
    ConfigParam("author", "Author(s) of the project"),
    ConfigParam("copyright_year", "Year when the project started"),
    ConfigParam("copyright_holder", "Copyright holder"),
    ConfigParam("project_url", "GitHub project url"),
    ConfigParam("bot_name", "Bot's GitHub login to push commits in CI"),
    ConfigParam("bot_email", "Bot account email"),
    ConfigParam("default_branch", "Default branch of the project"),
    ConfigParam("description", "Short description of the project"),
    ConfigParam(
        "license",
        "Project license type",
        type=click.Choice(["MIT", "Apache-2.0", "GPL-3.0", "None"], case_sensitive=False),
    ),
    MultiChoiceParam(
        "python_versions",
        text="Supported python versions",
        choices=["3.6", "3.7", "3.8", "3.9"],
    ),
    ConfigParam("docker_image", "Base docker image for the project's Docker container"),
    BooleanParam("docstring_checks", "Enable/disable linting docstrings"),
    ConfigParam("pyproject_extra", "Additional pyproject.toml configuration"),
    ConfigParam("ci_python_version", "Primary Python version in GitHub Actions"),
    ConfigParam("ci_ubuntu_version", "Primary Ubuntu version in GitHub Actions"),
    ConfigParam("ci_extra", "Additional script in GitHub Actions before runnign the main tests"),
    ConfigParam("git_message", "Initial Git commit message?"),
]

PROJECT, TEMPLATE = (
    NamedTuple(name, [(param.name, type(param)) for param in params])(*params)
    for (name, params) in [("PROJECT", _PROJECT), ("TEMPLATE", _TEMPLATE)]
)



"""Implements the `mloq setup` CLI command."""
import sys

import click
from omegaconf import DictConfig

from mloq.api import docs_project
from mloq.config.docs_generation import generate_config
from mloq.config.logic import write_config_docs
from mloq.config.params import TEMPLATE
from mloq.failure import Failure
from mloq.version import __version__


def generate_config_interactive(config: DictConfig) -> DictConfig:
    """Interactive generation of the project configuration."""
    # General project information
    click.echo("The following values will occur in several places in the generated files.")
    project, template = config.project, config.template
    template.project_name = TEMPLATE.project_name(template, True)
    template.description = TEMPLATE.description(template, True)
    template.author = TEMPLATE.author(template, True)
    click.echo()
    copyright_holder = TEMPLATE.copyright_holder(template, True, default=template.author)
    template.copyright_holder = copyright_holder
    click.echo()
    # Python version and requirements
    project.docs = True
    return config


def welcome_message():
    """Welcome message to be displayed during interactive documentation setup."""
    click.echo(f"Welcome to the MLOQ {__version__} interactive setup utility.")
    click.echo("This generates the necessary documentation for your new project.")
    click.echo()
    click.echo(
        "Please enter values for the following settings (just press Enter "
        "to accept a default value, if one is given in brackets).",
    )
    click.echo()


def docs_cmd(
    config: DictConfig,
    output,
    overwrite: bool,
    interactive: bool,
    only_config: bool,
) -> int:
    """
    Initialize the documentation of a new project using ML Ops Quickstart.

    Create the necessary documentation files of the new project (following
    ML Ops best practices).

    Args:
        config: Configuration Dictionary including the defined parameters.
        output: TODO
        overwrite: If True, overwrites the existent `mloq.yml` file.
        interactive: If True, activates the interactive mode to request
            the configuration inputs.
        only_config: TODO

    Returns:
        Returns an integer value indicating whether the process has
            ended successfully.
    """
    if interactive:
        welcome_message()
        config = generate_config_interactive(config)
    else:
        config = generate_config(config)
    if interactive and (only_config or click.confirm("Do you want to generate a mloq.yml file?")):
        write_config_docs(config, output, safe=True)
    if only_config:
        return 0
    if not overwrite and interactive:
        overwrite = click.confirm("Do you want to overwrite existing files?")
    try:
        docs_project(
            path=output,
            config=config,
            overwrite=overwrite,
        )
    except Failure as e:
        print(f"Failed to setup the project: {e}", file=sys.stderr)
        return 1
    return 0
